"""
Test the app settings in local.py
"""

# Django
from django.test import override_settings

# AA Sovereignty Timer
from sovtimer.app_settings import debug_enabled
from sovtimer.tests import BaseTestCase


class TestAppSettings(BaseTestCase):
    """
    Tests for App Settings
    """

    @override_settings(DEBUG=True)
    def test_debug_enabled_with_debug_true(self) -> None:
        """
        Test debug_enabled with DEBUG = True

        :return:
        :rtype:
        """

        self.assertTrue(debug_enabled())

    @override_settings(DEBUG=False)
    def test_debug_enabled_with_debug_false(self) -> None:
        """
        Test debug_enabled with DEBUG = False

        :return:
        :rtype:
        """

        self.assertFalse(debug_enabled())



# Generated by Django 4.2.14 on 2024-07-16 05:58

# Django
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("sovtimer", "0009_alter_campaign_options_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="campaign",
            name="event_type",
            field=models.CharField(
                choices=[
                    ("ihub_defense", "Sov Hub defense"),
                    ("tcu_defense", "TCU defense"),
                ],
                max_length=12,
            ),
        ),
    ]



# Generated by Django 3.1.3 on 2020-11-13 10:33

# Django
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("eveuniverse", "0004_effect_longer_name"),
        ("sovtimer", "0002_aasovtimercampaigns_aasovtimerstructures"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="aasovtimercampaigns",
            name="constellation_id",
        ),
        migrations.RemoveField(
            model_name="aasovtimercampaigns",
            name="defender_id",
        ),
        migrations.RemoveField(
            model_name="aasovtimercampaigns",
            name="solar_system_id",
        ),
        migrations.RemoveField(
            model_name="aasovtimerstructures",
            name="alliance_id",
        ),
        migrations.RemoveField(
            model_name="aasovtimerstructures",
            name="solar_system_id",
        ),
        migrations.AddField(
            model_name="aasovtimercampaigns",
            name="defender",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_DEFAULT,
                related_name="sov_campaign_defender",
                to="eveuniverse.eveentity",
            ),
        ),
        migrations.AddField(
            model_name="aasovtimercampaigns",
            name="solar_system",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_DEFAULT,
                related_name="sov_capaign_solar_system",
                to="eveuniverse.evesolarsystem",
            ),
        ),
        migrations.AddField(
            model_name="aasovtimerstructures",
            name="alliance",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_DEFAULT,
                related_name="sov_structure_alliance",
                to="eveuniverse.eveentity",
            ),
        ),
        migrations.AddField(
            model_name="aasovtimerstructures",
            name="solar_system",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_DEFAULT,
                related_name="sov_structure_solar_system",
                to="eveuniverse.evesolarsystem",
            ),
        ),
    ]



"""
Our Models
"""

# Django
from django.db import models
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag
from eveuniverse.models import EveEntity, EveSolarSystem

# AA Sovereignty Timer
from sovtimer import __title__
from sovtimer.handler import esi_handler
from sovtimer.providers import esi

logger = LoggerAddTag(my_logger=get_extension_logger(name=__name__), prefix=__title__)


class AaSovtimer(models.Model):
    """
    Meta model for app permissions
    """

    class Meta:
        """
        Meta definitions
        """

        verbose_name = _("Sovereignty Timer")
        managed = False
        default_permissions = ()
        permissions = (("basic_access", _("Can access the Sovereignty Timer module")),)


class SovereigntyStructure(models.Model):
    """
    Sov structures
    """

    structure_id = models.PositiveBigIntegerField(
        primary_key=True, db_index=True, unique=True
    )
    alliance = models.ForeignKey(
        to=EveEntity,
        on_delete=models.CASCADE,
        default=None,
        null=True,
        blank=True,
        related_name="sov_structure_alliance",
    )
    solar_system = models.ForeignKey(
        to=EveSolarSystem,
        on_delete=models.CASCADE,
        default=None,
        null=True,
        blank=True,
        related_name="sov_structure_solar_system",
    )
    structure_type_id = models.PositiveBigIntegerField()
    vulnerability_occupancy_level = models.FloatField(default=1)
    vulnerable_end_time = models.DateTimeField(null=True, blank=True)
    vulnerable_start_time = models.DateTimeField(null=True, blank=True)

    class Meta:
        """
        Meta definitions
        """

        verbose_name = _("Sovereignty structure")
        verbose_name_plural = _("Sovereignty structures")
        default_permissions = ()

    @classmethod
    def get_sov_structures_from_esi(cls, force_refresh: bool = False):
        """
        Get all sov structures from ESI

        :param force_refresh:
        :type force_refresh:
        :return:
        :rtype:
        """

        operation = esi.client.Sovereignty.GetSovereigntyStructures()
        sov_structures_from_esi = esi_handler.result(
            operation=operation, force_refresh=force_refresh
        )

        if sov_structures_from_esi:
            logger.debug(
                msg=f"Fetched {len(sov_structures_from_esi or [])} sovereignty structures from ESI."
            )
            logger.debug(msg=f"Sovereignty structure data: {sov_structures_from_esi}")
        else:
            logger.info(
                msg="No sovereignty structure changes found, nothing to update."
            )

        return sov_structures_from_esi


class Campaign(models.Model):
    """
    Sov campaigns
    """

    class Type(models.TextChoices):
        """
        Choices for Comment Types
        """

        SOVHUB_DEFENSE = "ihub_defense", _("Sov Hub defense")
        TCU_DEFENSE = "tcu_defense", _("TCU defense")

    campaign_id = models.PositiveBigIntegerField(
        primary_key=True, db_index=True, unique=True
    )
    attackers_score = models.FloatField(default=0.4)
    defender_score = models.FloatField(default=0.6)
    event_type = models.CharField(max_length=12, choices=Type.choices)
    start_time = models.DateTimeField()
    structure = models.OneToOneField(
        to=SovereigntyStructure,
        on_delete=models.CASCADE,
        default=None,
        null=True,
        blank=True,
        related_name="sov_capaign_structure",
    )

    progress_current = models.FloatField(default=0.6)
    progress_previous = models.FloatField(default=0.6)

    class Meta:
        """
        Meta definitions
        """

        verbose_name = _("Sovereignty campaign")
        verbose_name_plural = _("Sovereignty campaigns")
        default_permissions = ()

    @classmethod
    def get_sov_campaigns_from_esi(cls, force_refresh: bool = False):
        """
        Get all sov campaigns from ESI

        :return:
        """

        operation = esi.client.Sovereignty.GetSovereigntyCampaigns()
        campaigns_from_esi = esi_handler.result(
            operation=operation, force_refresh=force_refresh
        )

        if campaigns_from_esi:
            logger.debug(
                msg=f"Fetched {len(campaigns_from_esi or [])} campaigns from ESI."
            )
            logger.debug(msg=f"Campaign data: {campaigns_from_esi}")
        else:
            logger.info(msg="No campaign changes found, nothing to update.")

        return campaigns_from_esi



"""
App's URL config
"""

# Django
from django.urls import path

# AA Sovereignty Timer
from sovtimer import views
from sovtimer.constants import INTERNAL_URL_PREFIX

app_name: str = "sovtimer"  # pylint: disable=invalid-name

urlpatterns = [
    path(route="", view=views.dashboard, name="dashboard"),
    # Ajax call
    path(
        route=f"{INTERNAL_URL_PREFIX}/ajax/sov-campaign_data/",
        view=views.dashboard_data,
        name="dashboard_data",
    ),
]



"""
App settings
"""

# Standard Library
from re import RegexFlag

# Django
from django.conf import settings


def debug_enabled() -> RegexFlag:
    """
    Check if DEBUG is enabled

    :return:
    :rtype:
    """

    return settings.DEBUG



"""
Unit tests for the Campaign model's interaction with ESI.
"""

# Standard Library
from unittest.mock import MagicMock, patch

# Third Party
from aiopenapi3 import ContentTypeError

# Alliance Auth
from esi.exceptions import HTTPClientError, HTTPNotModified

# AA Sovereignty Timer
from sovtimer.models import Campaign
from sovtimer.tests import BaseTestCase


class CampaignESITestCase(BaseTestCase):
    """
    Test suite for the Campaign model's interaction with ESI.
    """

    def test_gets_sov_campaigns_from_esi_with_valid_data(self):
        """
        Test fetching sovereignty campaigns from ESI with valid data.

        :return:
        :rtype:
        """

        mock_esi_client = MagicMock()
        mock_esi_client.Sovereignty.GetSovereigntyCampaigns.return_value.result.return_value = [
            {"campaign_id": 1},
            {"campaign_id": 2},
        ]

        with patch("sovtimer.models.esi") as mock_esi_module:
            mock_esi_module.client = mock_esi_client
            result = Campaign.get_sov_campaigns_from_esi()

            self.assertEqual(result, [{"campaign_id": 1}, {"campaign_id": 2}])

    def test_returns_none_when_no_changes_in_sov_campaigns(self):
        """
        Test handling of no changes in sovereignty campaigns from ESI.

        :return:
        :rtype:
        """

        mock_esi_client = MagicMock()
        mock_esi_client.Sovereignty.GetSovereigntyCampaigns.return_value.result.side_effect = HTTPNotModified(
            status_code=304, headers={}
        )

        with patch("sovtimer.models.esi") as mock_esi_module:
            mock_esi_module.client = mock_esi_client
            result = Campaign.get_sov_campaigns_from_esi()

            self.assertIsNone(result)

    def test_handles_content_type_error_in_sov_campaigns(self):
        """
        Test handling of ContentTypeError when fetching sovereignty campaigns from ESI.

        :return:
        :rtype:
        """

        mock_esi_client = MagicMock()
        mock_esi_client.Sovereignty.GetSovereigntyCampaigns.return_value.result.side_effect = ContentTypeError(
            operation="dummy_op",
            content_type="application/json",
            message="dummy message",
            response=None,
        )

        with patch("sovtimer.models.esi") as mock_esi_module:
            mock_esi_module.client = mock_esi_client
            result = Campaign.get_sov_campaigns_from_esi()

            self.assertIsNone(result)

    def test_handles_httpclienterror_error_in_sov_campaigns(self):
        """
        Test handling of HTTPClientError when fetching sovereignty campaigns from ESI.

        :return:
        :rtype:
        """

        mock_esi_client = MagicMock()
        mock_esi_client.Sovereignty.GetSovereigntyCampaigns.return_value.result.side_effect = HTTPClientError(
            "HTTPClientError", {}, {}
        )
        with patch("sovtimer.models.esi") as mock_esi_module:
            mock_esi_module.client = mock_esi_client
            result = Campaign.get_sov_campaigns_from_esi()
            self.assertIsNone(result)



"""
App init
"""

# Django
from django.utils.translation import gettext_lazy as _

__version__ = "3.1.0"
__title__ = "Sovereignty Timer"
__title_translated__ = _("Sovereignty Timer")

__esi_compatibility_date__ = "2025-11-06"

__package_name__ = "aa-sov-timer"
__app_name__ = "sovtimer"
__app_name_verbose__ = "AA Sovereignty Timer"

__github_url__ = f"https://github.com/ppfeufer/{__package_name__}"



# Generated by Django 4.0.10 on 2023-09-22 22:46

# Django
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("sovtimer", "0008_fix_campaign_attackers_score"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="campaign",
            options={
                "default_permissions": (),
                "verbose_name": "Sovereignty campaign",
                "verbose_name_plural": "Sovereignty campaigns",
            },
        ),
        migrations.AlterModelOptions(
            name="sovereigntystructure",
            options={
                "default_permissions": (),
                "verbose_name": "Sovereignty structure",
                "verbose_name_plural": "Sovereignty structures",
            },
        ),
        migrations.AlterField(
            model_name="campaign",
            name="event_type",
            field=models.CharField(
                choices=[
                    ("ihub_defense", "IHub defense"),
                    ("tcu_defense", "TCU defense"),
                ],
                max_length=12,
            ),
        ),
    ]



"""
Test basic integration
"""

# Third Party
from django_webtest import WebTest

# Django
from django.urls import reverse


class TestSovtimerUI(WebTest):
    """
    Test Sovtimer UI
    """

    def test_should_show_sovtimer_dashboard(self):
        # when
        page = self.app.get(url=reverse(viewname="sovtimer:dashboard"))

        # then
        self.assertTemplateUsed(response=page, template_name="sovtimer/dashboard.html")



"""
The tasks
"""

# Third Party
from celery import shared_task

# Django
from django.core.cache import cache
from django.db import transaction

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag
from eveuniverse.models import EveEntity, EveSolarSystem

# AA Sovereignty Timer
from sovtimer import __title__
from sovtimer.models import Campaign, SovereigntyStructure

logger = LoggerAddTag(my_logger=get_extension_logger(name=__name__), prefix=__title__)


ESI_ERROR_LIMIT = 50
ESI_TIMEOUT_ONCE_ERROR_LIMIT_REACHED = 60
ESI_SOV_STRUCTURES_CACHE_KEY = "sov_structures_cache"
ESI_MAX_RETRIES = 5

TASK_TIME_LIMIT = 600  # Stop after 10 minutes

# Params for all tasks
TASK_PRIORITY = 6
TASK_ONCE_ARGS = {"graceful": True}
TASK_ONCE = {"base": QueueOnce, "once": TASK_ONCE_ARGS}

TASK_DEFAULTS = {
    "time_limit": TASK_TIME_LIMIT,
    "max_retries": ESI_MAX_RETRIES,
    "default_retry_delay": 300,
}
TASK_DEFAULTS_ONCE = {**TASK_DEFAULTS, **TASK_ONCE}


@shared_task(**TASK_DEFAULTS_ONCE)
def run_sov_campaign_updates() -> None:
    """
    Update all sovereignty campaigns and structures

    This task is used to update all sovereignty campaigns and structures from ESI.
    It should be called by a periodic task, or in your `local.py` settings file via:

    ```python
    # AA Sovereignty Timer - https://github.com/ppfeufer/aa-sov-timer
    if "sovtimer" in INSTALLED_APPS:
        # Run campaign updates every 30 seconds
        CELERYBEAT_SCHEDULE["sovtimer.tasks.run_sov_campaign_updates"] = {
            "task": "sovtimer.tasks.run_sov_campaign_updates",
            "schedule": 30,
        }
    ```
    """

    logger.info(msg="Updating sovereignty structures and campaigns from ESI …")

    update_sov_structures.apply_async(priority=TASK_PRIORITY, once=TASK_ONCE_ARGS)
    update_sov_campaigns.apply_async(priority=TASK_PRIORITY, once=TASK_ONCE_ARGS)


@shared_task(**TASK_DEFAULTS_ONCE)
def update_sov_campaigns(force_refresh: bool = False) -> None:
    """
    Update sovereignty campaigns

    This task is used to update the sovereignty campaigns from ESI.
    """

    campaigns_from_esi = Campaign.get_sov_campaigns_from_esi(
        force_refresh=force_refresh
    )

    if campaigns_from_esi is not None:
        logger.debug(
            msg=f"Number of sovereignty campaigns from ESI: {len(campaigns_from_esi or [])}"
        )

        campaigns = []
        defender_ids = {campaign.defender_id for campaign in campaigns_from_esi}

        existing_campaigns = {c.campaign_id: c for c in Campaign.objects.all()}

        EveEntity.objects.bulk_create(
            [EveEntity(id=defender_id) for defender_id in defender_ids],
            ignore_conflicts=True,
        )

        for campaign in campaigns_from_esi:
            campaign_id = campaign.campaign_id
            campaign_current__defender_score = campaign.defender_score
            campaign_current__progress_previous = campaign_current__defender_score

            if campaign_id in existing_campaigns:
                campaign_previous = existing_campaigns[campaign_id]
                campaign_previous__progress_previous = (
                    campaign_previous.progress_previous
                )
                campaign_previous__progress = campaign_previous.defender_score
                campaign_current__progress_previous = campaign_previous__progress

                if campaign_previous__progress == campaign_current__defender_score:
                    campaign_current__progress_previous = (
                        campaign_previous__progress_previous
                    )

            campaigns.append(
                Campaign(
                    attackers_score=campaign.attackers_score,
                    campaign_id=campaign_id,
                    defender_score=campaign.defender_score,
                    event_type=campaign.event_type,
                    start_time=campaign.start_time,
                    structure_id=campaign.structure_id,
                    progress_current=campaign_current__defender_score,
                    progress_previous=campaign_current__progress_previous,
                )
            )

        Campaign.objects.all().delete()
        Campaign.objects.bulk_create(
            objs=campaigns,
            batch_size=500,
            # ignore_conflicts=True,
        )
        EveEntity.objects.bulk_update_new_esi()

        logger.info(msg=f"{len(campaigns)} sovereignty campaigns updated from ESI.")


@shared_task(**TASK_DEFAULTS_ONCE)
def update_sov_structures(force_refresh: bool = False) -> None:
    """
    Update sovereignty structures

    This task is used to update the sovereignty structures from ESI.
    """

    # Check if the cache indicates that the structures have already been updated
    if not force_refresh and cache.get(ESI_SOV_STRUCTURES_CACHE_KEY):
        return

    structures_from_esi = SovereigntyStructure.get_sov_structures_from_esi(
        force_refresh=force_refresh
    )

    if structures_from_esi is not None:
        logger.debug(
            msg=f"Number of sovereignty structures from ESI: {len(structures_from_esi or [])}"
        )

        # Pre-fetch current structures and campaigns for fast lookup
        current_structures = {
            s["structure_id"]: s["vulnerability_occupancy_level"]
            for s in SovereigntyStructure.objects.all().values(
                "structure_id", "vulnerability_occupancy_level"
            )
        }

        # Pre-fetch EveEntities and SolarSystems to avoid repeated DB hits
        alliance_ids = {s.alliance_id for s in structures_from_esi if s.alliance_id}

        # Ensure we have EveEntity entries for all alliances
        EveEntity.objects.bulk_create(
            [EveEntity(id=aid) for aid in alliance_ids],
            ignore_conflicts=True,
        )

        # Fetch existing solar systems and alliances from the database
        solar_system_ids = {
            s.solar_system_id for s in structures_from_esi if s.solar_system_id
        }

        # Ensure we have EveSolarSystem entries for all solar systems
        solar_systems = {
            ss.id: ss for ss in EveSolarSystem.objects.filter(id__in=solar_system_ids)
        }

        # Ensure we have EveEntity entries for all alliances
        alliances = {e.id: e for e in EveEntity.objects.filter(id__in=alliance_ids)}

        esi_structure_ids = set()
        sov_structures = []

        # Iterate through the structures from ESI and prepare them for bulk creation
        for structure in structures_from_esi:
            structure_id = structure.structure_id

            # Skip structures without an ID or if they already exist in the set
            if not structure_id or structure_id in esi_structure_ids:
                continue

            esi_structure_ids.add(structure_id)

            sov_holder = alliances.get(structure.alliance_id)
            structure_solar_system = solar_systems.get(structure.solar_system_id)

            # Get the vulnerability occupancy level
            if structure_id in set(
                Campaign.objects.all().values_list("structure_id", flat=True)
            ):
                vulnerability_occupancy_level = current_structures.get(structure_id, 1)
            else:
                vulnerability_occupancy_level = (
                    structure.vulnerability_occupancy_level or 1
                )

            # Append the structure to the list for bulk creation
            sov_structures.append(
                SovereigntyStructure(
                    alliance=sov_holder,
                    solar_system=structure_solar_system,
                    structure_id=structure_id,
                    structure_type_id=structure.structure_type_id,
                    vulnerability_occupancy_level=vulnerability_occupancy_level,
                    vulnerable_end_time=structure.vulnerable_end_time,
                    vulnerable_start_time=structure.vulnerable_start_time,
                )
            )

        # Create or update the sovereignty structures in bulk
        with transaction.atomic():
            SovereigntyStructure.objects.bulk_create(
                objs=sov_structures,
                batch_size=500,
                update_conflicts=True,
                update_fields=[
                    "alliance",
                    "solar_system",
                    "structure_type_id",
                    "vulnerability_occupancy_level",
                    "vulnerable_end_time",
                    "vulnerable_start_time",
                ],
            )

            # Update EveEntity objects with new ESI data
            EveEntity.objects.bulk_update_new_esi()

            # Delete structures that are no longer present in ESI
            SovereigntyStructure.objects.exclude(pk__in=esi_structure_ids).delete()

        # Set the cache to indicate that the structures have been updated
        cache.set(key=ESI_SOV_STRUCTURES_CACHE_KEY, value=True, timeout=100)

        logger.info(
            msg=f"{len(sov_structures)} sovereignty structures updated from ESI."
        )



"""
ESI Handler for AA Sovereignty Timer
"""

# Standard Library
from typing import Any

# Third Party
from aiopenapi3 import ContentTypeError
from httpx import Response

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger
from esi.exceptions import HTTPClientError, HTTPNotModified
from esi.openapi_clients import EsiOperation

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Sovereignty Timer
from sovtimer import __title__

# Initialize a logger with a custom tag for the AA SRP application
logger = LoggerAddTag(my_logger=get_extension_logger(__name__), prefix=__title__)


def result(  # pylint: disable=too-many-arguments too-many-positional-arguments
    operation: EsiOperation,
    use_etag: bool = True,
    return_response: bool = False,
    force_refresh: bool = False,
    use_cache: bool = True,
    return_cached_for_304: bool = False,
    **extra,
) -> tuple[Any, Response] | Any:
    """
    Retrieve the result of an ESI operation, handling HTTPNotModified exceptions.

    :param operation: The ESI operation to execute.
    :type operation: EsiOperation
    :param use_etag: Whether to use ETag for caching.
    :type use_etag: bool
    :param return_response: Whether to return the full response object.
    :type return_response: bool
    :param force_refresh: Whether to force a refresh of the data.
    :type force_refresh: bool
    :param use_cache: Whether to use cached data.
    :type use_cache: bool
    :param extra: Additional parameters to pass to the operation.
    :param return_cached_for_304: Whether to return cached data on 304 Not Modified. If False, returns None for 304 responses.
    :type return_cached_for_304: bool
    :type extra: dict
    :return: The result of the ESI operation, optionally with the response object.
    :rtype: tuple[Any, Response] | Any
    """

    logger.debug(f"Handling ESI operation: {operation.operation.operationId}")

    try:
        esi_result = operation.result(
            use_etag=use_etag,
            return_response=return_response,
            force_refresh=force_refresh,
            use_cache=use_cache,
            **extra,
        )
    except HTTPNotModified:
        logger.debug(
            f"ESI returned 304 Not Modified for operation: {operation.operation.operationId} "
            f"- {'Using cached data.' if return_cached_for_304 else 'Skipping update.'}"
        )

        esi_result = operation.result(use_etag=False) if return_cached_for_304 else None
    except ContentTypeError:
        logger.warning(
            msg="ESI returned gibberish (ContentTypeError), skipping campaign update."
        )

        esi_result = None
    except HTTPClientError as exc:
        logger.error(msg=f"Error while fetching campaigns from ESI: {str(exc)}")

        esi_result = None

    return esi_result



"""
Loading initial data into sovtimer tables
"""

# Django
from django.core.management.base import BaseCommand

# Alliance Auth (External Libs)
from eveuniverse.core.esitools import is_esi_online

# AA Sovereignty Timer
from sovtimer.tasks import update_sov_campaigns, update_sov_structures

ESI_SOV_STRUCTURES_CACHE_KEY = "sov_structures_cache"


def get_input(text):
    """
    Get input from the user
    """

    return input(text)


class Command(BaseCommand):
    """
    Imports initial sov campaign data
    """

    help = "Imports initial data"

    def _import_sov_data(self) -> None:  # pylint: disable=too-many-locals
        """
        Import sovereignty data

        :return:
        """

        if not is_esi_online():
            self.stdout.write(
                msg="ESI is currently offline. Can not start ESI related tasks. Aborting"
            )

            return

        update_sov_structures(force_refresh=True)
        update_sov_campaigns(force_refresh=True)

    def handle(self, *args, **options):  # pylint: disable=unused-argument
        """
        Ask before running …

        :param args:
        :param options:
        """

        self.stdout.write(
            msg=(
                "This will start the initial import for SOV campaigns and structures. "
                "It's quite a bit to import, so this might take a moment or two. "
                "Please be patient …"
            )
        )

        user_input = get_input(text="Are you sure you want to proceed? (yes/no)?")

        if user_input == "yes":
            self.stdout.write(msg="Starting import. Please stand by.")
            self._import_sov_data()
            self.stdout.write(msg=self.style.SUCCESS("Import complete!"))
        else:
            self.stdout.write(msg=self.style.WARNING("Aborted."))



"""
Unit tests for the ESI handler functions in the sovtimer application.
"""

# Standard Library
from http import HTTPStatus
from unittest.mock import MagicMock

# Third Party
from aiopenapi3 import ContentTypeError

# Alliance Auth
from esi.exceptions import HTTPClientError, HTTPNotModified

# AA Sovereignty Timer
from sovtimer.handler.esi_handler import result
from sovtimer.tests import BaseTestCase


class TestHandlerEsi(BaseTestCase):
    """
    Test ESI handler functions
    """

    def test_returns_result_when_operation_succeeds(self):
        """
        Test that the result function returns the expected data when the operation succeeds.

        :return:
        :rtype:
        """

        op = MagicMock()
        op.operation.operationId = "test_operation"
        op.result.return_value = {"data": "success"}

        res = result(op)

        self.assertEqual(res, {"data": "success"})

    def test_returns_cached_data_when_http_not_modified_and_cached_enabled(self):
        """
        Test that the result function returns cached data when an HTTPNotModified exception is raised and caching is enabled.

        :return:
        :rtype:
        """

        op = MagicMock()
        op.operation.operationId = "test_operation"
        op.result.side_effect = [
            HTTPNotModified(HTTPStatus.NOT_MODIFIED, {}),
            {"data": "cached"},
        ]

        res = result(op, return_cached_for_304=True)

        self.assertEqual(res, {"data": "cached"})

    def test_returns_none_when_http_not_modified_and_cached_disabled(self):
        """
        Test that the result function returns None when an HTTPNotModified exception is raised and caching is disabled.

        :return:
        :rtype:
        """

        op = MagicMock()
        op.operation.operationId = "test_operation"
        op.result.side_effect = HTTPNotModified(HTTPStatus.NOT_MODIFIED, {})

        res = result(op, return_cached_for_304=False)

        self.assertIsNone(res)

    def test_returns_none_when_content_type_error_occurs(self):
        """
        Test that the result function returns None when a ContentTypeError exception is raised.

        :return:
        :rtype:
        """

        op = MagicMock()
        op.operation.operationId = "test_operation"
        op.result.side_effect = ContentTypeError(
            op.operation, "text/plain", "Invalid content type", MagicMock()
        )

        res = result(op)

        self.assertIsNone(res)

    def test_returns_none_when_http_client_error_occurs(self):
        """
        Test that the result function returns None when an HTTPClientError exception is raised.

        :return:
        :rtype:
        """

        op = MagicMock()
        op.operation.operationId = "test_operation"
        op.result.side_effect = HTTPClientError(
            HTTPStatus.BAD_REQUEST, {}, b"Client error"
        )

        res = result(op)

        self.assertIsNone(res)



"""
App config
"""

# Django
from django.apps import AppConfig
from django.utils.text import format_lazy

# AA Sovereignty Timer
from sovtimer import __title_translated__, __version__


class AaSovtimerConfig(AppConfig):
    """
    Application config
    """

    name = "sovtimer"
    label = "sovtimer"
    verbose_name = format_lazy(
        "{app_title} v{version}", app_title=__title_translated__, version=__version__
    )



# Generated by Django 3.1.3 on 2020-11-13 18:56

# Django
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("sovtimer", "0003_auto_20201113_1033"),
    ]

    operations = [
        migrations.AddField(
            model_name="aasovtimercampaigns",
            name="progress_current",
            field=models.FloatField(default=0.6),
        ),
        migrations.AddField(
            model_name="aasovtimercampaigns",
            name="progress_previous",
            field=models.FloatField(default=0.6),
        ),
        migrations.AlterField(
            model_name="aasovtimercampaigns",
            name="attackers_score",
            field=models.FloatField(default=0.6),
        ),
        migrations.AlterField(
            model_name="aasovtimercampaigns",
            name="defender_score",
            field=models.FloatField(default=0.6),
        ),
    ]



"""
Tests for the sovtimer tasks.
"""

# Standard Library
from types import SimpleNamespace
from unittest.mock import MagicMock, Mock, patch

# Django
from django.utils import timezone

# Alliance Auth (External Libs)
from eveuniverse.models import EveSolarSystem

# AA Sovereignty Timer
from sovtimer.models import Campaign, SovereigntyStructure
from sovtimer.tasks import (
    TASK_ONCE_ARGS,
    TASK_PRIORITY,
    run_sov_campaign_updates,
    update_sov_campaigns,
    update_sov_structures,
)
from sovtimer.tests import BaseTestCase


class TestRunSovCampaignUpdatesTask(BaseTestCase):
    """
    Test the run_sov_campaign_updates task.
    """

    @patch("sovtimer.tasks.update_sov_structures.apply_async")
    @patch("sovtimer.tasks.update_sov_campaigns.apply_async")
    def test_run_sov_campaign_updates_calls_tasks(
        self, mock_update_campaigns, mock_update_structures
    ):
        """
        Test that run_sov_campaign_updates calls the update tasks with correct args.

        :param mock_update_campaigns:
        :type mock_update_campaigns:
        :param mock_update_structures:
        :type mock_update_structures:
        :return:
        :rtype:
        """

        run_sov_campaign_updates()

        mock_update_structures.assert_called_once_with(
            priority=TASK_PRIORITY, once=TASK_ONCE_ARGS
        )
        mock_update_campaigns.assert_called_once_with(
            priority=TASK_PRIORITY, once=TASK_ONCE_ARGS
        )

    def test_run_sov_campaign_updates_handles_exceptions(self):
        """
        Test that run_sov_campaign_updates handles exceptions gracefully.

        :return:
        :rtype:
        """

        with (
            patch(
                "sovtimer.tasks.update_sov_structures.apply_async"
            ) as mock_update_structures,
            patch("sovtimer.tasks.update_sov_campaigns.apply_async"),
        ):
            mock_update_structures.side_effect = Exception("Test exception")

            with self.assertRaises(Exception) as exc:
                run_sov_campaign_updates()

            self.assertEqual(str(exc.exception), "Test exception")

    def test_run_sov_campaign_updates_logs_messages(self):
        """
        Test that run_sov_campaign_updates logs the correct messages.

        :return:
        :rtype:
        """

        with (
            patch("sovtimer.tasks.logger") as mock_logger,
            patch("sovtimer.tasks.update_sov_structures.apply_async"),
            patch("sovtimer.tasks.update_sov_campaigns.apply_async"),
        ):
            run_sov_campaign_updates()

            mock_logger.info.assert_any_call(
                msg="Updating sovereignty structures and campaigns from ESI …"
            )


class TestUpdateSovCampaignsTask(BaseTestCase):
    """
    Test the update_sov_campaigns task.
    """

    @patch("sovtimer.models.Campaign.get_sov_campaigns_from_esi")
    @patch("sovtimer.models.Campaign.objects.bulk_create")
    @patch("sovtimer.models.Campaign.objects.all")
    @patch("sovtimer.models.EveEntity.objects.bulk_create")
    @patch("sovtimer.models.EveEntity.objects.bulk_update_new_esi")
    @patch("sovtimer.tasks.logger.info")
    @patch("sovtimer.tasks.logger.debug")
    def test_updates_campaigns_and_logs_info(
        self,
        mock_logger_debug,
        mock_logger_info,
        mock_bulk_update_new_esi,
        mock_eve_bulk_create,
        mock_campaigns_all,
        mock_bulk_create,
        mock_get_campaigns,
    ):
        """
        Test that update_sov_campaigns correctly updates campaigns and logs info.

        :param mock_logger_debug:
        :type mock_logger_debug:
        :param mock_logger_info:
        :type mock_logger_info:
        :param mock_bulk_update_new_esi:
        :type mock_bulk_update_new_esi:
        :param mock_eve_bulk_create:
        :type mock_eve_bulk_create:
        :param mock_campaigns_all:
        :type mock_campaigns_all:
        :param mock_bulk_create:
        :type mock_bulk_create:
        :param mock_get_campaigns:
        :type mock_get_campaigns:
        :return:
        :rtype:
        """

        mock_queryset = Mock()
        mock_queryset.__iter__ = lambda self: iter(
            [Mock(campaign_id=1), Mock(campaign_id=2)]
        )
        mock_campaigns_all.return_value = mock_queryset

        mock_get_campaigns.return_value = [
            Mock(
                campaign_id=1,
                defender_id=1001,
                attackers_score=0.5,
                defender_score=0.7,
                event_type="ihub_defense",
                start_time="2023-10-01T12:00:00Z",
                structure_id=2001,
            ),
            Mock(
                campaign_id=2,
                defender_id=1002,
                attackers_score=0.3,
                defender_score=0.6,
                event_type="ihub_defense",
                start_time="2023-10-02T12:00:00Z",
                structure_id=2002,
            ),
        ]

        update_sov_campaigns()

        mock_logger_debug.assert_called_with(
            msg="Number of sovereignty campaigns from ESI: 2"
        )
        mock_logger_info.assert_called_with(
            msg="2 sovereignty campaigns updated from ESI."
        )
        mock_eve_bulk_create.assert_called_once()
        mock_bulk_create.assert_called_once()
        mock_bulk_update_new_esi.assert_called_once()

    @patch("sovtimer.models.Campaign.get_sov_campaigns_from_esi")
    @patch("sovtimer.models.Campaign.objects.all")
    @patch("sovtimer.tasks.logger.info")
    @patch("sovtimer.tasks.logger.debug")
    def test_handles_no_campaigns_from_esi(
        self,
        mock_logger_debug,
        mock_logger_info,
        mock_campaigns_all,
        mock_get_campaigns,
    ):
        """
        Test that update_sov_campaigns handles no campaigns returned from ESI.

        :param mock_logger_debug:
        :type mock_logger_debug:
        :param mock_logger_info:
        :type mock_logger_info:
        :param mock_campaigns_all:
        :type mock_campaigns_all:
        :param mock_get_campaigns:
        :type mock_get_campaigns:
        :return:
        :rtype:
        """

        mock_get_campaigns.return_value = ""

        update_sov_campaigns()

        mock_logger_debug.assert_called_with(
            msg="Number of sovereignty campaigns from ESI: 0"
        )
        mock_logger_info.assert_called_with(
            msg="0 sovereignty campaigns updated from ESI."
        )
        mock_campaigns_all.return_value.delete.assert_called_once()

    @patch("sovtimer.models.Campaign.get_sov_campaigns_from_esi")
    @patch("sovtimer.models.Campaign.objects.all")
    @patch("sovtimer.models.Campaign.objects.bulk_create")
    @patch("sovtimer.models.EveEntity.objects.bulk_create")
    @patch("sovtimer.models.EveEntity.objects.bulk_update_new_esi")
    @patch("sovtimer.tasks.logger.info")
    @patch("sovtimer.tasks.logger.debug")
    def test_deletes_existing_campaigns_when_new_campaigns_arrive(
        self,
        mock_logger_debug,
        mock_logger_info,
        mock_bulk_update_new_esi,
        mock_eve_bulk_create,
        mock_bulk_create,
        mock_campaigns_all,
        mock_get_campaigns,
    ):
        mock_queryset = Mock()
        mock_queryset.__iter__ = lambda self: iter([Mock(campaign_id=1)])
        mock_queryset.delete = Mock()
        mock_campaigns_all.return_value = mock_queryset

        mock_get_campaigns.return_value = [
            Mock(
                campaign_id=1,
                defender_id=1001,
                attackers_score=0.5,
                defender_score=0.7,
                event_type="ihub_defense",
                start_time="2023-10-01T12:00:00Z",
                structure_id=2001,
            )
        ]

        update_sov_campaigns()

        mock_campaigns_all.return_value.delete.assert_called_once()
        mock_bulk_create.assert_called_once()
        mock_eve_bulk_create.assert_called_once()
        mock_bulk_update_new_esi.assert_called_once()
        mock_logger_info.assert_called_with(
            msg="1 sovereignty campaigns updated from ESI."
        )

    @patch("sovtimer.models.EveEntity.objects.bulk_create")
    @patch("sovtimer.models.EveEntity.objects.bulk_update_new_esi")
    def test_preserves_previous_progress_previous_when_defender_score_unchanged(
        self, mock_bulk_update_new_esi, mock_eve_bulk_create
    ):
        now = timezone.now()
        # Create the referenced SovereigntyStructure so FK constraint is satisfied
        SovereigntyStructure.objects.create(
            structure_id=100,
            alliance_id=None,
            solar_system_id=None,
            structure_type_id=0,
            vulnerability_occupancy_level=0,
            vulnerable_start_time=now,
            vulnerable_end_time=now,
        )

        # Existing campaign in DB with previous progress_previous
        Campaign.objects.create(
            attackers_score=0,
            campaign_id=1,
            defender_score=50,
            event_type="test",
            start_time=now,
            structure_id=100,
            progress_current=50,
            progress_previous=20,
        )

        esi_campaign = SimpleNamespace(
            attackers_score=0,
            campaign_id=1,
            defender_id=1001,
            defender_score=50,  # unchanged
            event_type="test",
            start_time=now,
            structure_id=100,
        )

        with patch(
            "sovtimer.tasks.Campaign.get_sov_campaigns_from_esi",
            return_value=[esi_campaign],
        ):
            update_sov_campaigns(force_refresh=True)

        created = Campaign.objects.get(campaign_id=1)
        self.assertEqual(created.progress_current, 50)
        self.assertEqual(created.progress_previous, 20)

    @patch("sovtimer.models.EveEntity.objects.bulk_create")
    @patch("sovtimer.models.EveEntity.objects.bulk_update_new_esi")
    def test_sets_progress_previous_to_current_defender_score_when_defender_score_changed(
        self, mock_bulk_update_new_esi, mock_eve_bulk_create
    ):
        now = timezone.now()
        # Create the referenced SovereigntyStructure so FK constraint is satisfied
        SovereigntyStructure.objects.create(
            structure_id=200,
            alliance_id=None,
            solar_system_id=None,
            structure_type_id=0,
            vulnerability_occupancy_level=0,
            vulnerable_start_time=now,
            vulnerable_end_time=now,
        )

        # Existing campaign in DB with a different defender_score
        Campaign.objects.create(
            attackers_score=0,
            campaign_id=2,
            defender_score=40,
            event_type="test",
            start_time=now,
            structure_id=200,
            progress_current=40,
            progress_previous=10,
        )

        esi_campaign = SimpleNamespace(
            attackers_score=0,
            campaign_id=2,
            defender_id=1002,
            defender_score=60,  # changed
            event_type="test",
            start_time=now,
            structure_id=200,
        )

        with patch(
            "sovtimer.tasks.Campaign.get_sov_campaigns_from_esi",
            return_value=[esi_campaign],
        ):
            update_sov_campaigns(force_refresh=True)

        created = Campaign.objects.get(campaign_id=2)
        self.assertEqual(created.progress_current, 60)
        self.assertEqual(created.progress_previous, 40)

    @patch("sovtimer.models.EveEntity.objects.bulk_create")
    @patch("sovtimer.models.EveEntity.objects.bulk_update_new_esi")
    def test_sets_progress_previous_to_current_when_no_existing_campaign(
        self, mock_bulk_update_new_esi, mock_eve_bulk_create
    ):
        now = timezone.now()
        Campaign.objects.all().delete()

        # Create the referenced SovereigntyStructure so FK constraint is satisfied
        SovereigntyStructure.objects.create(
            structure_id=300,
            alliance_id=None,
            solar_system_id=None,
            structure_type_id=0,
            vulnerability_occupancy_level=0,
            vulnerable_start_time=now,
            vulnerable_end_time=now,
        )

        esi_campaign = SimpleNamespace(
            attackers_score=0,
            campaign_id=3,
            defender_id=1003,
            defender_score=30,
            event_type="test",
            start_time=now,
            structure_id=300,
        )

        with patch(
            "sovtimer.tasks.Campaign.get_sov_campaigns_from_esi",
            return_value=[esi_campaign],
        ):
            update_sov_campaigns(force_refresh=True)

        created = Campaign.objects.get(campaign_id=3)
        self.assertEqual(created.progress_current, 30)
        self.assertEqual(created.progress_previous, 30)


class TestUpdateSovStructuresTask(BaseTestCase):
    """
    Test the update_sov_structures task.
    """

    @patch("sovtimer.models.SovereigntyStructure.get_sov_structures_from_esi")
    @patch("sovtimer.models.SovereigntyStructure.objects.bulk_create")
    @patch("sovtimer.models.SovereigntyStructure.objects.exclude")
    @patch("sovtimer.models.SovereigntyStructure.objects.all")
    @patch("sovtimer.models.EveEntity.objects.bulk_create")
    @patch("sovtimer.models.EveEntity.objects.bulk_update_new_esi")
    @patch("sovtimer.models.EveSolarSystem.objects.filter")
    @patch("sovtimer.tasks.cache.get")
    @patch("sovtimer.tasks.cache.set")
    @patch("sovtimer.tasks.logger.info")
    @patch("sovtimer.tasks.logger.debug")
    def test_updates_structures_and_logs_info(
        self,
        mock_logger_debug,
        mock_logger_info,
        mock_cache_set,
        mock_cache_get,
        mock_solar_system_filter,
        mock_bulk_update_new_esi,
        mock_eve_bulk_create,
        mock_structures_all,
        mock_structures_exclude,
        mock_bulk_create,
        mock_get_structures,
    ):
        """
        Test that update_sov_structures correctly updates structures and logs info.

        :param mock_logger_debug:
        :type mock_logger_debug:
        :param mock_logger_info:
        :type mock_logger_info:
        :param mock_cache_set:
        :type mock_cache_set:
        :param mock_cache_get:
        :type mock_cache_get:
        :param mock_solar_system_filter:
        :type mock_solar_system_filter:
        :param mock_bulk_update_new_esi:
        :type mock_bulk_update_new_esi:
        :param mock_eve_bulk_create:
        :type mock_eve_bulk_create:
        :param mock_structures_all:
        :type mock_structures_all:
        :param mock_structures_exclude:
        :type mock_structures_exclude:
        :param mock_bulk_create:
        :type mock_bulk_create:
        :param mock_get_structures:
        :type mock_get_structures:
        :return:
        :rtype:
        """

        mock_cache_get.return_value = None
        mock_get_structures.return_value = [
            Mock(
                structure_id=1,
                alliance_id=1001,
                solar_system_id=2001,
                structure_type_id=3001,
                vulnerability_occupancy_level=0.5,
                vulnerable_start_time="2023-10-01T12:00:00Z",
                vulnerable_end_time="2023-10-02T12:00:00Z",
            )
        ]
        mock_structures_all.return_value.values.return_value = [
            {"structure_id": 1, "vulnerability_occupancy_level": 0.5}
        ]

        mock_solar_system = MagicMock(spec=EveSolarSystem)
        mock_solar_system.id = 2001
        mock_solar_system._state = MagicMock()  # Add the _state attribute
        mock_solar_system_filter.return_value = [mock_solar_system]

        mock_structures_exclude.return_value.delete = Mock()

        update_sov_structures()

        mock_logger_debug.assert_called_with(
            msg="Number of sovereignty structures from ESI: 1"
        )
        mock_logger_info.assert_called_with(
            msg="1 sovereignty structures updated from ESI."
        )
        mock_eve_bulk_create.assert_called_once()
        mock_bulk_create.assert_called_once()
        mock_bulk_update_new_esi.assert_called_once()
        mock_cache_set.assert_called_once()

    @patch("sovtimer.models.SovereigntyStructure.get_sov_structures_from_esi")
    @patch("sovtimer.tasks.cache.get")
    def test_skips_update_when_cache_is_set(self, mock_cache_get, mock_get_structures):
        """
        Test that update_sov_structures skips update when cache is set.

        :param mock_cache_get:
        :type mock_cache_get:
        :param mock_get_structures:
        :type mock_get_structures:
        :return:
        :rtype:
        """

        mock_cache_get.return_value = True

        update_sov_structures()

        mock_get_structures.assert_not_called()

    @patch("sovtimer.models.SovereigntyStructure.get_sov_structures_from_esi")
    @patch("sovtimer.tasks.logger.info")
    @patch("sovtimer.tasks.logger.debug")
    def test_handles_no_structures_from_esi(
        self, mock_logger_debug, mock_logger_info, mock_get_structures
    ):
        mock_get_structures.return_value = None

        update_sov_structures()

        mock_logger_debug.assert_not_called()
        mock_logger_info.assert_not_called()

    def test_ignores_duplicate_structures_with_same_id(self):
        """
        Test that update_sov_structures ignores duplicate structures with the same ID.

        :return:
        :rtype:
        """

        now = timezone.now()
        struct_a = SimpleNamespace(
            structure_id=55555,
            alliance_id=None,
            solar_system_id=None,
            structure_type_id=3000,
            vulnerability_occupancy_level=1,
            vulnerable_end_time=now,
            vulnerable_start_time=now,
        )
        struct_b = SimpleNamespace(
            structure_id=55555,  # duplicate ID
            alliance_id=None,
            solar_system_id=None,
            structure_type_id=3001,
            vulnerability_occupancy_level=3,
            vulnerable_end_time=now,
            vulnerable_start_time=now,
        )

        with patch(
            "sovtimer.models.SovereigntyStructure.get_sov_structures_from_esi",
            return_value=[struct_a, struct_b],
        ):
            update_sov_structures(force_refresh=True)

        # Only one SovereigntyStructure with that structure_id should exist
        self.assertEqual(
            SovereigntyStructure.objects.filter(structure_id=55555).count(), 1
        )

    def test_returns_default_vulnerability_level_when_structure_not_in_campaign(self):
        """
        Test that the default vulnerability level is returned when structure is not in campaign.

        :return:
        :rtype:
        """

        structure_id = 99999
        current_structures = {88888: 2}

        with patch.object(Campaign.objects, "all", autospec=True) as mock_all:
            mock_all.return_value.values_list.return_value = [(12345,), (67890,)]

            result = current_structures.get(structure_id, 1)

            self.assertEqual(result, 1)

    def test_returns_existing_vulnerability_level_when_structure_in_campaign(self):
        """
        Test that the existing vulnerability level is returned when structure is in campaign.

        :return:
        :rtype:
        """

        structure_id = 12345
        current_structures = {12345: 3}

        with patch.object(Campaign.objects, "all", autospec=True) as mock_all:
            mock_all.return_value.values_list.return_value = [(12345,), (67890,)]

            result = current_structures.get(structure_id, 1)

            self.assertEqual(result, 3)



"""
Test ajax calls
"""

# Standard Library
import json
from http import HTTPStatus

# Django
from django.test import RequestFactory
from django.urls import reverse

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter

# Alliance Auth (External Libs)
from app_utils.testing import add_character_to_user, create_user_from_evecharacter

# AA Sovereignty Timer
from sovtimer.tests import BaseTestCase
from sovtimer.tests.fixtures.load_allianceauth import load_allianceauth
from sovtimer.tests.fixtures.load_sovtimer import load_sovtimer
from sovtimer.views import dashboard_data


class TestAjaxCalls(BaseTestCase):
    """
    Test the ajax calls
    """

    @classmethod
    def setUpClass(cls):
        """
        Setup

        :return:
        :rtype:
        """

        super().setUpClass()
        load_allianceauth()

        cls.factory = RequestFactory()

        # given
        cls.character_1001 = EveCharacter.objects.get(character_id=1001)
        cls.character_1002 = EveCharacter.objects.get(character_id=1002)
        cls.character_1003 = EveCharacter.objects.get(character_id=1003)
        cls.character_1004 = EveCharacter.objects.get(character_id=1004)
        cls.character_1005 = EveCharacter.objects.get(character_id=1005)
        cls.character_1101 = EveCharacter.objects.get(character_id=1101)

        cls.user_with_basic_access, _ = create_user_from_evecharacter(
            character_id=cls.character_1001.character_id,
            permissions=["sovtimer.basic_access"],
        )

        add_character_to_user(
            user=cls.user_with_basic_access, character=cls.character_1101
        )

        cls.user_without_access, _ = create_user_from_evecharacter(
            character_id=cls.character_1002.character_id
        )

        add_character_to_user(
            user=cls.user_without_access, character=cls.character_1003
        )

    def test_ajax_dashboard_data_no_campaigns(self):
        # given
        self.client.force_login(user=self.user_with_basic_access)
        request = self.factory.get(path=reverse(viewname="sovtimer:dashboard"))
        request.user = self.user_with_basic_access

        # when
        result = dashboard_data(request=request)

        # then
        self.assertEqual(first=result.status_code, second=HTTPStatus.OK)
        self.assertJSONEqual(raw=result.content, expected_data=[])

    def test_ajax_dashboard_data_with_campaigns(self):
        """
        Test the ajax call to get the dashboard data with campaigns

        :return:
        :rtype:
        """

        # given
        load_sovtimer()
        self.client.force_login(user=self.user_with_basic_access)
        request = self.factory.get(path=reverse(viewname="sovtimer:dashboard"))
        request.user = self.user_with_basic_access

        # when
        result = dashboard_data(request=request)
        response_data = json.loads(result.content)

        # Check structure and types separately
        self.assertEqual(first=result.status_code, second=HTTPStatus.OK)
        self.assertEqual(len(response_data), 3)

        # Check first campaign has positive remaining time
        self.assertIsInstance(
            response_data[0]["remaining_time_in_seconds"], (int, float)
        )
        self.assertGreater(response_data[0]["remaining_time_in_seconds"], 0)
        self.assertEqual(response_data[0]["campaign_progress"], "60%")

        # Check second campaign (attackers making progress) has negative remaining time (past event)
        self.assertIsInstance(
            response_data[1]["remaining_time_in_seconds"], (int, float)
        )
        self.assertLess(response_data[1]["remaining_time_in_seconds"], 0)
        self.assertIn(
            '13%<i class="material-icons aa-sovtimer-trend aa-sovtimer-trend-down" title="Attackers making progress" data-bs-tooltip="aa-sovtimer">trending_down</i>04%',
            response_data[1]["campaign_progress"],
        )

        # Check third campaign has (defenders making progress) negative remaining time (past event)
        self.assertIsInstance(
            response_data[2]["remaining_time_in_seconds"], (int, float)
        )
        self.assertLess(response_data[2]["remaining_time_in_seconds"], 0)
        self.assertIn(
            '04%<i class="material-icons aa-sovtimer-trend aa-sovtimer-trend-up" title="Defenders making progress" data-bs-tooltip="aa-sovtimer">trending_up</i>13%',
            response_data[2]["campaign_progress"],
        )

        # Check other fields remain exact
        expected_fields = {
            "event_type": "TCU defense",
            "solar_system_name": "0-O6XF",
            "defender_name": "Wayne Enterprises",
            "adm": 6.0,
        }

        for field, value in expected_fields.items():
            self.assertEqual(response_data[0][field], value)
            self.assertEqual(response_data[1][field], value)
            self.assertEqual(response_data[2][field], value)



# Generated by Django 3.2.11 on 2022-01-30 11:36

# Django
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("eveuniverse", "0005_type_materials_and_sections"),
        ("sovtimer", "0006_rename_models"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="campaign",
            name="defender",
        ),
        migrations.RemoveField(
            model_name="campaign",
            name="solar_system",
        ),
        migrations.RemoveField(
            model_name="campaign",
            name="structure_id",
        ),
        migrations.AddField(
            model_name="campaign",
            name="structure",
            field=models.OneToOneField(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="sov_capaign_structure",
                to="sovtimer.sovereigntystructure",
            ),
        ),
        migrations.AlterField(
            model_name="sovereigntystructure",
            name="alliance",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="sov_structure_alliance",
                to="eveuniverse.eveentity",
            ),
        ),
        migrations.AlterField(
            model_name="sovereigntystructure",
            name="solar_system",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="sov_structure_solar_system",
                to="eveuniverse.evesolarsystem",
            ),
        ),
        migrations.AlterField(
            model_name="sovereigntystructure",
            name="vulnerability_occupancy_level",
            field=models.FloatField(default=1),
        ),
        migrations.AlterField(
            model_name="sovereigntystructure",
            name="vulnerable_end_time",
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="sovereigntystructure",
            name="vulnerable_start_time",
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]



"""
Hook into AA
"""

# Alliance Auth
from allianceauth import hooks
from allianceauth.services.hooks import MenuItemHook, UrlHook

# AA Sovereignty Timer
from sovtimer import __title_translated__, urls


class AaSovtimerMenuItem(MenuItemHook):  # pylint: disable=too-few-public-methods
    """
    This class ensures only authorized users will see the menu entry
    """

    def __init__(self):
        # Setup menu entry for sidebar
        MenuItemHook.__init__(
            self,
            text=__title_translated__,
            classes="fa-regular fa-clock",
            url_name="sovtimer:dashboard",
            navactive=["sovtimer:"],
        )

    def render(self, request):
        """
        Check if the user has the permission to view this app

        :param request:
        :return:
        """

        return MenuItemHook.render(self, request=request)


@hooks.register("menu_item_hook")
def register_menu():
    """
    Register our menu item

    :return:
    """

    return AaSovtimerMenuItem()


@hooks.register("url_hook")
def register_urls():
    """
    Register our base url

    :return:
    """

    return UrlHook(
        urls=urls,
        namespace="sovtimer",
        base_url=r"^sovereignty-timer/",
        excluded_views=["sovtimer.views.dashboard", "sovtimer.views.dashboard_data"],
    )
